import inspect
from inspect import getmro
from typing import Any, Callable, Dict, List, Mapping, Optional, Type, Union, cast

from esmerald.enums import MediaType, ScopeType
from esmerald.exception_handlers import http_exception_handler
from esmerald.exceptions import HTTPException, WebSocketException
from esmerald.requests import Request
from esmerald.responses import Response
from esmerald.types import ExceptionHandler, ExceptionHandlers
from esmerald.websockets import WebSocket
from pydantic import BaseModel
from starlette import status
from starlette._utils import is_async_callable
from starlette.concurrency import run_in_threadpool
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.middleware.exceptions import (
    ExceptionMiddleware as StarletteExceptionMiddleware,
)
from starlette.responses import Response as StarletteResponse
from starlette.types import ASGIApp, Message, Receive, Scope, Send


class ExceptionMiddleware(StarletteExceptionMiddleware):
    """
    Reimplementation of the Exception Middleware.
    """

    def __init__(
        self,
        app: ASGIApp,
        handlers: Optional[Mapping[Any, Callable[[Request, Exception], Response]]] = None,
        debug: bool = False,
    ) -> None:
        self.app = app
        self.debug = debug
        self._status_handlers: Dict[int, Callable] = {}
        self._exception_handlers: Dict[Type[Exception], Callable] = {
            HTTPException: http_exception_handler,
            StarletteHTTPException: http_exception_handler,
            WebSocketException: self.websocket_exception,
        }
        if handlers is not None:
            for key, value in handlers.items():
                self.add_exception_handler(key, value)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        response_started = False

        async def sender(message: Message) -> None:
            nonlocal response_started

            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        try:
            await self.app(scope=scope, receive=receive, send=send)
        except Exception as exc:
            handler = None

            if isinstance(exc, StarletteHTTPException):
                handler = self._status_handlers.get(exc.status_code)

            if handler is None:
                handler = self._lookup_exception_handler(exc)

            if handler is None:
                raise exc

            if response_started:
                msg = "Caught handled exception, but response already started."
                raise RuntimeError(msg) from exc

            if scope["type"] == "http":
                request = Request(scope, receive=receive)
                if is_async_callable(handler):
                    response = await handler(request, exc)
                else:
                    response = await run_in_threadpool(handler, request, exc)
                await response(scope, receive, sender)
            elif scope["type"] == "websocket":
                websocket = WebSocket(scope, receive=receive, send=send)
                if is_async_callable(handler):
                    if inspect.isfunction(handler):
                        await self.app(scope, receive, send)
                    else:
                        await handler(websocket, exc)
                else:
                    await run_in_threadpool(handler, websocket, exc)


class ResponseContent(BaseModel):
    detail: Optional[str]
    extra: Optional[Union[Dict[str, Any], List[Any]]] = None
    headers: Optional[Dict[str, str]] = None
    status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR


class EsmeraldAPIExceptionMiddleware:
    def __init__(
        self,
        app: "ASGIApp",
        debug: bool,
        exception_handlers: "ExceptionHandlers",
        error_handler: Optional[Callable] = None,
    ) -> None:
        self.app = app
        self.exception_handlers = exception_handlers
        self.debug = debug
        self.error_handler = error_handler

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        try:
            await self.app(scope, receive, send)
        except Exception as ex:
            if scope["type"] == ScopeType.HTTP:
                exception_handler = (
                    self.get_exception_handler(self.exception_handlers, ex)
                    or self.default_http_exception_handler
                )
                response = exception_handler(Request(scope, receive, send), ex)
                await response(scope, receive, send)
                return

            if isinstance(ex, WebSocketException):
                code = ex.code
                reason = ex.detail
            elif isinstance(ex, StarletteHTTPException):
                code = ex.status_code + 4000
                reason = ex.detail
            else:
                code = status.HTTP_500_INTERNAL_SERVER_ERROR + 4000
                reason = repr(ex)

            event = {"type": "websocket.close", "code": code, "reason": reason}
            await send(event)

    def default_http_exception_handler(
        self, request: Request, exc: Exception
    ) -> "StarletteResponse":
        """Default handler for exceptions subclassed from HTTPException."""
        status_code = (
            exc.status_code
            if isinstance(exc, StarletteHTTPException)
            else status.HTTP_500_INTERNAL_SERVER_ERROR
        )
        if status_code == status.HTTP_500_INTERNAL_SERVER_ERROR and self.debug:
            server_error = ServerErrorMiddleware(app=self.app, handler=self.error_handler)  # type: ignore[arg-type]
            return server_error.debug_response(request=request, exc=exc)  # type: ignore[arg-type]
        return self.create_exception_response(exc)

    def create_exception_response(self, exc: Exception) -> Response:
        if isinstance(exc, (HTTPException, StarletteHTTPException)):
            content = ResponseContent(detail=exc.detail, status_code=exc.status_code)
            if isinstance(exc, HTTPException):
                extra = exc.extra.get("extra", {})
                if extra:
                    content.extra = extra
        else:
            content = ResponseContent(detail=repr(exc))
        return Response(
            media_type=MediaType.JSON,
            content=content.dict(exclude_none=True),
            status_code=content.status_code,
            headers=exc.headers
            if isinstance(exc, (HTTPException, StarletteHTTPException))
            else None,
        )

    def get_exception_handler(
        self,
        exception_handlers: Dict[Union[int, Type[Exception]], "ExceptionHandlers"],
        exc: Exception,
    ) -> Optional[ExceptionHandler]:
        if not exception_handlers:
            return None

        if not isinstance(exc, HTTPException) and exc.__class__ in exception_handlers:
            return exception_handlers[exc.__class__]

        if isinstance(exc, HTTPException) and exc.__class__ in exception_handlers:
            return exception_handlers[exc.__class__]

        for klass in getmro(type(exc)):
            if klass in exception_handlers:
                return exception_handlers[cast("Type[Exception]", exc)]
        return None



from typing import Union

from asyncz.triggers import (
    AndTrigger,
    CronTrigger,
    DateTrigger,
    IntervalTrigger,
    OrTrigger,
)

TriggerHandler = Union[
    AndTrigger,
    CronTrigger,
    DateTrigger,
    IntervalTrigger,
    OrTrigger,
]

TriggerType = Union[str, TriggerHandler]



import asyncio
import functools
import sys
import typing
from inspect import isclass
from typing import Any, Awaitable, Callable, TypeVar, Union

import slugify
from typing_extensions import ParamSpec, TypeGuard, get_args, get_origin

if sys.version_info >= (3, 10):
    from types import UnionType

    UNION_TYPES = {UnionType, Union}
else:  # pragma: no cover
    UNION_TYPES = {Union}

P = ParamSpec("P")
T = TypeVar("T")


def is_async_callable(value: Callable[P, T]) -> TypeGuard[Callable[P, Awaitable[T]]]:
    while isinstance(value, functools.partial):
        value = value.func  # type: ignore[unreachable]

    return asyncio.iscoroutinefunction(value) or asyncio.iscoroutinefunction(value.__call__)  # ty


def is_class_and_subclass(value: typing.Any, _type: typing.Any) -> bool:
    original = get_origin(value)
    if not original and not isclass(value):
        return False

    try:
        if original:
            return original and issubclass(original, _type)
        return issubclass(value, _type)
    except TypeError:
        return False


def clean_string(value: str) -> str:
    return slugify.slugify(value, separator="_")


def is_optional_union(annotation: Any) -> bool:
    return get_origin(annotation) in UNION_TYPES and type(None) in get_args(annotation)



from typing import TYPE_CHECKING, Any, Dict, Optional, Set, Tuple, Type, Union

from esmerald.enums import EncodingType, ParamType
from esmerald.exceptions import ImproperlyConfigured
from esmerald.parsers import BaseModelExtra, parse_form_data
from esmerald.requests import Request
from esmerald.transformers.datastructures import EsmeraldSignature as SignatureModel
from esmerald.transformers.utils import (
    Dependency,
    ParamSetting,
    create_parameter_setting,
    get_request_params,
    get_signature,
    merge_sets,
)
from esmerald.utils.constants import RESERVED_KWARGS
from esmerald.utils.pydantic import is_field_optional
from pydantic.fields import (
    SHAPE_DEQUE,
    SHAPE_FROZENSET,
    SHAPE_LIST,
    SHAPE_SEQUENCE,
    SHAPE_SET,
    SHAPE_TUPLE,
    SHAPE_TUPLE_ELLIPSIS,
    ModelField,
)

if TYPE_CHECKING:
    from esmerald.types import Dependencies, ReservedKwargs
    from esmerald.websockets import WebSocket
    from pydantic.typing import DictAny


MEDIA_TYPES = [EncodingType.MULTI_PART, EncodingType.URL_ENCODED]


class TransformerModel(BaseModelExtra):
    class Config(BaseModelExtra.Config):
        arbitrary_types_allowed = True

    def __init__(
        self,
        cookies: Set[ParamSetting],
        dependencies: Set[Dependency],
        form_data: Optional[Tuple[EncodingType, ModelField]],
        headers: Set[ParamSetting],
        path_params: Set[ParamSetting],
        query_params: Set[ParamSetting],
        reserved_kwargs: Set["ReservedKwargs"],
        query_param_names: Set[str],
        is_optional: bool,
        **kwargs: "DictAny",
    ):
        super().__init__(**kwargs)
        self.cookies = cookies
        self.dependencies = dependencies
        self.form_data = form_data
        self.headers = headers
        self.path_params = path_params
        self.query_params = query_params
        self.reserved_kwargs = reserved_kwargs
        self.query_param_names = query_param_names
        self.has_kwargs = (
            cookies
            or dependencies
            or form_data
            or headers
            or path_params
            or query_params
            or reserved_kwargs
        )
        self.is_optional = is_optional

    @classmethod
    def dependency_tree(cls, key: str, dependencies: "Dependencies") -> Dependency:
        inject = dependencies[key]
        dependency_keys = [key for key in get_signature(inject).__fields__ if key in dependencies]
        return Dependency(
            key=key,
            inject=inject,
            dependencies=[
                cls.dependency_tree(key=key, dependencies=dependencies) for key in dependency_keys
            ],
        )

    @classmethod
    def get_parameter_settings(
        cls,
        path_parameters: Set[str],
        dependencies: "Dependencies",
        signature_fields: Dict[str, ModelField],
    ) -> Tuple[Set[ParamSetting], set]:
        shapes = {
            SHAPE_LIST,
            SHAPE_SET,
            SHAPE_SEQUENCE,
            SHAPE_TUPLE,
            SHAPE_TUPLE_ELLIPSIS,
            SHAPE_DEQUE,
            SHAPE_FROZENSET,
        }

        _dependencies = set()

        for key in dependencies:
            if key in signature_fields:
                _dependencies.add(cls.dependency_tree(key=key, dependencies=dependencies))

        ignored_keys = {*RESERVED_KWARGS, *(dependency.key for dependency in _dependencies)}

        parameter_definitions = set()
        for field_name, model_field in signature_fields.items():
            if field_name not in ignored_keys:
                parameter_definitions.add(
                    create_parameter_setting(
                        allow_none=model_field.allow_none,
                        field_name=field_name,
                        field_info=model_field.field_info,
                        path_parameters=path_parameters,
                        is_sequence=model_field.shape in shapes,
                    )
                )

        for field_name, model_field in filter(
            lambda items: items[0] not in ignored_keys,
            signature_fields.items(),
        ):
            signature_field = model_field.field_info
            parameter_definitions.add(
                create_parameter_setting(
                    allow_none=model_field.allow_none,
                    field_name=field_name,
                    field_info=signature_field,
                    path_parameters=path_parameters,
                    is_sequence=model_field.shape in shapes,
                )
            )

        return parameter_definitions, _dependencies

    @classmethod
    def create_signature(
        cls,
        signature_model: Type[SignatureModel],
        dependencies: "Dependencies",
        path_parameters: Set[str],
    ) -> "TransformerModel":
        cls.validate_kwargs(
            path_parameters=path_parameters,
            dependencies=dependencies,
            model_fields=signature_model.__fields__,
        )

        reserved_kwargs = set()

        for field_name in signature_model.__fields__:
            if field_name in RESERVED_KWARGS:
                reserved_kwargs.add(field_name)

        param_settings, _dependencies = cls.get_parameter_settings(
            path_parameters=path_parameters,
            dependencies=dependencies,
            signature_fields=signature_model.__fields__,
        )

        path_params = set()
        for param in param_settings:
            if param.param_type == ParamType.PATH:
                path_params.add(param)

        headers = set()
        for param in param_settings:
            if param.param_type == ParamType.HEADER:
                headers.add(param)

        cookies = set()
        for param in param_settings:
            if param.param_type == ParamType.COOKIE:
                cookies.add(param)

        query_params = set()
        for param in param_settings:
            if param.param_type == ParamType.QUERY:
                query_params.add(param)

        query_params_names = set()
        for param in param_settings:
            if param.param_type == ParamType.QUERY and param.is_sequence:
                query_params_names.add(param)

        form_data = None

        # For the reserved keyword data
        data_field = signature_model.__fields__.get("data")
        if data_field:
            media_type = data_field.field_info.extra.get("media_type")
            if media_type in MEDIA_TYPES:
                form_data = (media_type, data_field)

        path_params, query_params, cookies, headers, reserved_kwargs = cls.update_parameters(
            global_dependencies=dependencies,
            local_dependencies=_dependencies,
            path_params=path_params,
            query_params=query_params,
            cookies=cookies,
            headers=headers,
            reserved_kwargs=reserved_kwargs,
            path_parameters=path_parameters,
            form_data=form_data,
        )

        is_optional = False
        if "data" in reserved_kwargs:
            is_optional = is_field_optional(signature_model.__fields__["data"])

        return TransformerModel(
            form_data=form_data,
            dependencies=_dependencies,
            path_params=path_params,
            query_params=query_params,
            cookies=cookies,
            headers=headers,
            reserved_kwargs=reserved_kwargs,
            query_param_names=query_params_names,
            is_optional=is_optional,
        )

    @classmethod
    def update_parameters(
        cls,
        global_dependencies: "Dependencies",
        local_dependencies: Set["Dependency"],
        path_params: "DictAny",
        query_params: "DictAny",
        cookies: "DictAny",
        headers: "DictAny",
        reserved_kwargs: "DictAny",
        path_parameters: "DictAny",
        form_data: "DictAny",
    ) -> Tuple["DictAny", "DictAny", "DictAny", "DictAny", "DictAny"]:
        for dependency in local_dependencies:
            dependency_model = cls.create_signature(
                signature_model=get_signature(dependency.inject),
                dependencies=global_dependencies,
                path_parameters=path_parameters,
            )
            path_params = merge_sets(path_params, dependency_model.path_params)
            query_params = merge_sets(query_params, dependency_model.query_params)
            cookies = merge_sets(cookies, dependency_model.cookies)
            headers = merge_sets(headers, dependency_model.headers)

            if "data" in reserved_kwargs and "data" in dependency_model.reserved_kwargs:
                cls.validate_data(form_data, dependency_model)
            reserved_kwargs.update(dependency_model.reserved_kwargs)

        return path_params, query_params, cookies, headers, reserved_kwargs

    def to_kwargs(self, connection: Union["WebSocket", "Request"]) -> "DictAny":
        connection_params = {}
        for key, value in connection.query_params.items():
            if key not in self.query_param_names and len(value) == 1:
                value = value[0]
                connection_params[key] = value

        query_params = get_request_params(
            params=connection.query_params, expected=self.query_params, url=connection.url
        )
        path_params = get_request_params(
            params=connection.path_params, expected=self.path_params, url=connection.url
        )
        headers = get_request_params(
            params=connection.headers, expected=self.headers, url=connection.url
        )
        cookies = get_request_params(
            params=connection.cookies, expected=self.cookies, url=connection.url
        )

        if not self.reserved_kwargs:
            return {**query_params, **path_params, **headers, **cookies}

        return self.handle_reserved_kwargs(
            connection=connection,
            connection_params=connection_params,
            path_params=path_params,
            query_params=query_params,
            headers=headers,
            cookies=cookies,
        )

    def handle_reserved_kwargs(
        self,
        connection: Union["WebSocket", "Request"],
        connection_params: "DictAny",
        path_params: "DictAny",
        query_params: "DictAny",
        headers: "DictAny",
        cookies: "DictAny",
    ) -> "DictAny":
        reserved_kwargs: "DictAny" = {}
        if "data" in self.reserved_kwargs:
            reserved_kwargs["data"] = self.get_request_data(request=connection)
        if "request" in self.reserved_kwargs:
            reserved_kwargs["request"] = connection
        if "socket" in self.reserved_kwargs:
            reserved_kwargs["socket"] = connection
        if "headers" in self.reserved_kwargs:
            reserved_kwargs["headers"] = connection.headers
        if "cookies" in self.reserved_kwargs:
            reserved_kwargs["cookies"] = connection.cookies
        if "query" in self.reserved_kwargs:
            reserved_kwargs["query"] = connection_params
        if "state" in self.reserved_kwargs:
            reserved_kwargs["state"] = connection.app.state.copy()

        return {**reserved_kwargs, **path_params, **query_params, **headers, **cookies}

    @classmethod
    def validate_data(
        cls,
        form_data: Optional[Tuple[EncodingType, ModelField]],
        dependency_model: "TransformerModel",
    ) -> None:
        if form_data and dependency_model.form_data:
            media_type, _ = form_data
            dependency_media_type, _ = dependency_model.form_data
            if media_type != dependency_media_type:
                raise ImproperlyConfigured(
                    "Dependencies have incompatible form-data encoding. "
                    "They should both be the same. Either url-encoded or multi-part."
                )
        if (
            (form_data and not dependency_model.form_data)
            or not form_data
            and dependency_model.form_data
        ):
            raise ImproperlyConfigured(
                "Dependencies haev incompativle 'data' kwarg types. "
                "One expects JSON and the other expects form-data."
            )

    @classmethod
    def validate_kwargs(
        cls,
        path_parameters: Set[str],
        dependencies: "Dependencies",
        model_fields: Dict[str, ModelField],
    ) -> None:
        keys = set(dependencies.keys())
        names = set()

        for key, value in model_fields.items():
            if (
                value.field_info.extra.get(ParamType.QUERY)
                or value.field_info.extra.get(ParamType.HEADER)
                or value.field_info.extra.get(ParamType.COOKIE)
            ):
                names.add(key)

        for intersect in [
            path_parameters.intersection(keys)
            or path_parameters.intersection(names)
            or keys.intersection(names)
        ]:
            if intersect:
                raise ImproperlyConfigured(
                    f"Ambiguity in kwarg resolution: {', '.join(intersect)}. "
                    f"Please make sure to use unique distinct keys for your dependencies and path parameters."
                )

        reserved_kwargs = {*names, *path_parameters, *keys}.intersection(RESERVED_KWARGS)
        if reserved_kwargs:
            raise ImproperlyConfigured(
                f"Kwargs ({', '.join(RESERVED_KWARGS)}) cannot be used for parameters and/or dependencies."
            )

    async def get_request_data(self, request: "Request") -> Any:
        # Fast exit principle
        if not self.form_data:
            return await request.json()

        media_type, field = self.form_data
        form_data = await request.form()
        parsed_form = parse_form_data(media_type, form_data, field)
        return parsed_form if parsed_form or not self.is_optional else None

    async def get_dependencies(
        self,
        dependency: "Dependency",
        connection: Union["WebSocket", "Request"],
        **kwargs: "DictAny",
    ) -> Any:
        signature_model = get_signature(dependency.inject)
        for _dependency in dependency.dependencies:
            kwargs[_dependency.key] = await self.get_dependencies(
                dependency=_dependency, connection=connection, **kwargs
            )
        dependency_kwargs = signature_model.parse_values_for_connection(
            connection=connection, **kwargs
        )
        return await dependency.inject(**dependency_kwargs)



from copy import copy
from typing import TYPE_CHECKING, List, Optional, Union, cast

from starlette.routing import compile_path

from esmerald.utils.url import clean_path

if TYPE_CHECKING:
    from esmerald.permissions.types import Permission
    from esmerald.routing.router import HTTPHandler, Router, WebSocketHandler
    from esmerald.types import (
        Dependencies,
        ExceptionHandlers,
        Middleware,
        ResponseCookies,
        ResponseHeaders,
        ResponseType,
    )


class APIView:
    """The Esmerald APIView class.

    Subclassing this class will create a view using Class Based Views.
    """

    __slots__ = (
        "dependencies",
        "exception_handlers",
        "interceptors",
        "permissions",
        "middleware",
        "parent",
        "path",
        "response_class",
        "response_cookies",
        "response_headers",
        "tags",
        "path_regex",
        "path_format",
        "param_convertors",
        "deprecated",
    )

    path: str
    dependencies: Optional["Dependencies"]
    exception_handlers: Optional["ExceptionHandlers"]
    permissions: Optional[List["Permission"]]
    middleware: Optional[List["Middleware"]]
    parent: "Router"
    response_class: Optional["ResponseType"]
    response_cookies: Optional["ResponseCookies"]
    response_headers: Optional["ResponseHeaders"]
    tags: Optional[List[str]]
    deprecated: Optional[bool]

    def __init__(self, parent: "Router") -> None:
        for key in self.__slots__:
            if not hasattr(self, key):
                setattr(self, key, None)

        self.path = clean_path(self.path or "/")
        self.path_regex, self.path_format, self.param_convertors = compile_path(self.path)
        self.parent = parent
        self.interceptors = []

    def get_filtered_handler(self) -> List[str]:
        """
        Filters out the names of the functions that are not part of the handler itself.
        """
        from esmerald.routing.router import HTTPHandler, WebSocketHandler

        filtered_handlers = [
            attr for attr in dir(self) if not attr.startswith("__") and not attr.endswith("__")
        ]
        route_handlers = []

        for handler_name in filtered_handlers:

            if handler_name not in dir(APIView) and isinstance(
                getattr(self, handler_name), (HTTPHandler, WebSocketHandler)
            ):
                route_handlers.append(handler_name)
        return route_handlers

    def get_route_handlers(self) -> List[Union["HTTPHandler", "WebSocketHandler"]]:
        """A getter for the apiview's route handlers that sets their parent.

        Returns:
            A list containing a copy of the route handlers defined inside the APIView.
        """
        route_handlers: List[Union["HTTPHandler", "WebSocketHandler"]] = []
        filtered_handlers = self.get_filtered_handler()

        for handler in filtered_handlers:
            source_route_handler = cast(
                Union["HTTPHandler", "WebSocketHandler"], getattr(self, handler)
            )
            route_handler = copy(source_route_handler)
            route_handler.parent = self

            if self.middleware:
                self.get_route_middleware(route_handler)

            if self.exception_handlers:
                route_handler.exception_handlers = self.get_exception_handlers(route_handler)
            if self.tags or []:
                for tag in reversed(self.tags):
                    route_handler.tags.insert(0, tag)
            route_handlers.append(route_handler)

        return route_handlers

    def get_route_middleware(
        self, handler: Union["HTTPHandler", "WebSocketHandler"]
    ) -> List["Middleware"]:
        """
        Gets the list of extended middlewares for the handler starting from the last
        to the first by reversing the list
        """
        for middleware in reversed(self.middleware):
            handler.middleware.insert(0, middleware)

    def get_exception_handlers(
        self, handler: Union["HTTPHandler", "WebSocketHandler"]
    ) -> List["ExceptionHandlers"]:
        """
        Gets the dict of extended exception handlers for the handler starting from the last
        to the first by reversing the list
        """
        exception_handlers = {**self.exception_handlers, **handler.exception_handlers}
        return exception_handlers



from http import HTTPStatus
from inspect import Signature
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Tuple, Type, cast

from esmerald.datastructures import File, Redirect, Stream, Template
from esmerald.enums import MediaType
from esmerald.exceptions import (
    HTTPException,
    ImproperlyConfigured,
    ValidationErrorException,
)
from esmerald.openapi.enums import OpenAPIFormat, OpenAPIType
from esmerald.openapi.schema import create_schema
from esmerald.openapi.utils import pascal_case_to_text
from esmerald.responses import Response as EsmeraldResponse
from esmerald.utils.model import create_parsed_model_field
from openapi_schemas_pydantic.v3_1_0 import Response
from openapi_schemas_pydantic.v3_1_0.header import Header
from openapi_schemas_pydantic.v3_1_0.media_type import (
    MediaType as OpenAPISchemaMediaType,
)
from openapi_schemas_pydantic.v3_1_0.schema import Schema
from starlette.routing import get_name
from typing_extensions import get_args, get_origin

if TYPE_CHECKING:

    from esmerald.datastructures import Cookie
    from esmerald.routing.router import HTTPHandler
    from esmerald.types import AnyCallable
    from openapi_schemas_pydantic.v3_1_0.responses import Responses


def create_cookie_schema(cookie: "Cookie") -> Schema:
    cookie_copy = cookie.copy(update={"value": "<string>"})
    value = cookie_copy.to_header(header="")
    return Schema(description=cookie.description or "", example=value)


def create_success_response(handler: "HTTPHandler", create_examples: bool) -> Response:
    signature = Signature.from_callable(cast("AnyCallable", handler.fn))
    default_descriptions: Dict[Any, str] = {
        Stream: "Stream Response",
        Redirect: "Redirect Response",
        File: "File Download",
    }

    description = (
        handler.response_description
        or default_descriptions.get(signature.return_annotation)
        or HTTPStatus(handler.status_code).description
    )

    if signature.return_annotation not in {
        signature.empty,
        None,
        Redirect,
        File,
        Stream,
    }:
        return_annotation = signature.return_annotation
        if signature.return_annotation is Template:
            return_annotation = str
            handler.media_type = MediaType.HTML
        elif get_origin(signature.return_annotation) is EsmeraldResponse:
            return_annotation = get_args(signature.return_annotation)[0] or Any
        as_parsed_model_field = create_parsed_model_field(return_annotation)
        schema = create_schema(field=as_parsed_model_field, create_examples=create_examples)
        schema.contentEncoding = handler.content_encoding
        schema.contentMediaType = handler.content_media_type
        response = Response(
            content={handler.media_type: OpenAPISchemaMediaType(media_type_schema=schema)},
            description=description,
        )
    elif signature.return_annotation is Redirect:
        response = Response(
            content=None,
            description=description,
            headers={
                "location": Header(
                    param_schema=Schema(type=OpenAPIType.STRING),
                    description="target path for redirect",
                )
            },
        )
    elif signature.return_annotation in (File, Stream):
        response = Response(
            content={
                handler.media_type: OpenAPISchemaMediaType(
                    media_type_schema=Schema(
                        type=OpenAPIType.STRING,
                        contentEncoding=handler.content_encoding or "application/octet-stream",
                        contentMediaType=handler.content_media_type,
                    )
                )
            },
            description=description,
            headers={
                "content-length": Header(
                    param_schema=Schema(type=OpenAPIType.STRING),
                    description="File size in bytes",
                ),
                "last-modified": Header(
                    param_schema=Schema(
                        type=OpenAPIType.STRING, schema_format=OpenAPIFormat.DATE_TIME
                    ),
                    description="Last modified data-time in RFC 2822 format",
                ),
                "etag": Header(
                    param_schema=Schema(type=OpenAPIType.STRING),
                    description="Entity tag",
                ),
            },
        )
    else:
        response = Response(content=None, description=description)

    if response.headers is None:
        response.headers = {}

    for key, value in handler.get_response_headers().items():
        header = Header()
        for attribute_name, attribute_value in value.dict(exclude_none=True).items():
            if attribute_name == "value":
                model_field = create_parsed_model_field(type(attribute_value))
                header.param_schema = create_schema(field=model_field, create_examples=False)
        response.headers[key] = header
    cookies = handler.get_response_cookies()
    if cookies:
        response.headers["Set-Cookie"] = Header(
            param_schema=Schema(allOF=[create_cookie_schema(cookie=cookie) for cookie in cookies])
        )
    return response


def create_error_responses(
    exceptions: List[Type[HTTPException]],
) -> Iterator[Tuple[str, Response]]:

    grouped_exceptions: Dict[int, List[Type[HTTPException]]] = {}
    for exc in exceptions:
        if not grouped_exceptions.get(exc.status_code):
            grouped_exceptions[exc.status_code] = []
        grouped_exceptions[exc.status_code].append(exc)

    for status_code, exception_group in grouped_exceptions.items():
        exception_schemas = [
            Schema(
                type=OpenAPIType.OBJECT,
                required=["detail", "status_code"],
                properties=dict(
                    status_code=Schema(type=OpenAPIType.INTEGER),
                    detail=Schema(type=OpenAPIType.STRING),
                    extra=Schema(
                        type=[OpenAPIType.NULL, OpenAPIType.OBJECT, OpenAPIType.ARRAY],
                        additionalProperties=Schema(),
                    ),
                ),
                description=pascal_case_to_text((get_name(exc))),
                examples=[
                    {
                        "status_code": status_code,
                        "detail": HTTPStatus(status_code).phrase,
                        "extra": {},
                    }
                ],
            )
            for exc in exception_group
        ]

        if len(exception_schemas) > 1:
            schema = Schema(oneOf=exception_schemas)
        else:
            schema = exception_schemas[0]
        yield str(status_code), Response(
            description=HTTPStatus(status_code).description,
            content={MediaType.JSON: OpenAPISchemaMediaType(media_type_schema=schema)},
        )


def create_additional_responses(
    handler: "HTTPHandler",
) -> Iterator[Tuple[str, Response]]:
    if not handler.responses:
        return

    for status_code, additional_response in handler.responses.items():
        model_field = create_parsed_model_field(additional_response.model)
        schema = create_schema(
            field=model_field, create_examples=additional_response.create_examples
        )

        yield str(status_code), Response(
            description=additional_response.description,
            content={
                additional_response.media_type: OpenAPISchemaMediaType(media_type_schema=schema)
            },
        )


def create_responses(
    handler: "HTTPHandler", raises_validation_error: bool, create_examples: bool
) -> Optional["Responses"]:

    responses: "Responses" = {
        str(handler.status_code): create_success_response(
            handler=handler, create_examples=create_examples
        )
    }

    exceptions = handler.raise_exceptions or []
    if raises_validation_error and ValidationErrorException not in exceptions:
        exceptions.append(ValidationErrorException)

    for status_code, response in create_error_responses(exceptions):
        responses[status_code] = response

    for status_code, response in create_additional_responses(handler):
        if status_code in responses:
            raise ImproperlyConfigured(
                f"Additional response for status code {status_code} already exists in success or error responses"
            )
        responses[status_code] = response
    return responses or None



from typing import Optional

from esmerald.concurrency import AsyncExitStack
from esmerald.config import AsyncExitConfig
from esmerald.protocols.middleware import MiddlewareProtocol
from esmerald.types import ASGIApp, Receive, Scope, Send


class AsyncExitStackMiddleware(MiddlewareProtocol):
    def __init__(self, app: "ASGIApp", config: "AsyncExitConfig"):
        """AsyncExitStack Middleware class.

        Args:
            app: The 'next' ASGI app to call.
            config: The AsyncExitConfig instance.
        """
        super().__init__(app)
        self.app = app
        self.config = config

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        if not AsyncExitStack:
            await self.app(scope, receive, send)

        exception: Optional[Exception] = None
        async with AsyncExitStack() as stack:
            scope[self.config.context_name] = stack
            try:
                await self.app(scope, receive, send)
            except Exception as e:
                exception = e
                raise e
            if exception:
                raise exception



from typing import Type, TypeVar, Union

from esmerald.interceptors.interceptor import EsmeraldInterceptor

EsmeraldInterceptorType = TypeVar("EsmeraldInterceptorType", bound=EsmeraldInterceptor)
Interceptor = Union[Type[EsmeraldInterceptorType], EsmeraldInterceptor]



from os import PathLike
from typing import TYPE_CHECKING, Any, List, Union

from esmerald.exceptions import MissingDependency, TemplateNotFound
from esmerald.protocols.template import TemplateEngineProtocol

if TYPE_CHECKING:
    from pydantic import DirectoryPath

try:
    from jinja2 import Environment, FileSystemLoader
    from jinja2 import Template as JinjaTemplate
    from jinja2 import TemplateNotFound as JinjaTemplateNotFound
except ImportError as exc:
    raise MissingDependency("jinja2 is not installed") from exc


try:
    import jinja2

    # @contextfunction was renamed to @pass_context in Jinja 3.0, and was removed in 3.1
    # hence we try to get pass_context (most installs will be >=3.1)
    # and fall back to contextfunction,
    # adding a type ignore for mypy to let us access an attribute that may not exist
    if hasattr(jinja2, "pass_context"):
        pass_context = jinja2.pass_context
    else:  # pragma: nocover
        pass_context = jinja2.contextfunction  # type: ignore[attr-defined]
except ImportError:  # pragma: nocover
    jinja2 = None  # type: ignore


class JinjaTemplateEngine(TemplateEngineProtocol[JinjaTemplate]):
    def __init__(
        self, directory: Union["DirectoryPath", List["DirectoryPath"]], **env_options: Any
    ) -> None:
        super().__init__(directory)
        self.env = self._create_environment(directory, **env_options)

    def _create_environment(
        self, directory: Union[str, PathLike], **env_options: Any
    ) -> "jinja2.Environment":
        @pass_context
        def url_for(context: dict, name: str, **path_params: Any) -> str:
            request = context["request"]
            return request.url_for(name, **path_params)

        loader = FileSystemLoader(directory)
        env_options.setdefault("loader", loader)
        env_options.setdefault("autoescape", True)

        env = Environment(**env_options)
        env.globals["url_for"] = url_for
        return env

    def get_template(self, name: str) -> JinjaTemplate:
        return self.env.get_template(name)

    def _get_template(self, template_name: str) -> JinjaTemplate:
        try:
            return self.engine.get_template(name=template_name)
        except JinjaTemplateNotFound as e:
            raise TemplateNotFound(template_name=template_name) from e



from typing import TYPE_CHECKING, TypeVar

from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from starlette.types import Receive, Scope, Send

T = TypeVar("T")


@runtime_checkable
class InterceptorProtocol(Protocol):  # pragma: no cover
    """
    Generic object serving the base for interception of messages,
    before reaching the endpoint. This is inspired by the AOP (Aspect Oriented Programming).

    The interceptor is handled between the call and the API endpoint itself and acts on it.

    An interceptor could be anything from logging to rerouting or even input sanitizing.
    """

    async def intercept(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        ...



"""
Functions to use with the Router.
"""
from importlib import import_module
from typing import Any, Optional

from esmerald.exceptions import ImproperlyConfigured

DEFAULT_PATTERN = "route_patterns"


def include(arg: Any, pattern: Optional[str] = DEFAULT_PATTERN):
    """Simple retrieve functionality to make it easier to include
    routes in the urls. Example, nested routes.

    Example:

        # myapp.routes.py

        from esmerald import Router
        from .views import MyView

        router_patterns = [
            Router(path='/my-iew', routes=[MyView])
        ]

        # routers.py
        from esmerald import Router
        from esmerald.conf.urls import include

        router_patterns = [
            Router(path='/', routes=[include('myapp.routes')])
        ]

    If a `routes.py ` doesn't contain a default `router_patterns` that can be
    also specified dynamically.

    Example:

        # myapp.routes.py

        from esmerald import Router
        from .views import MyView

        myapp_urls = [
            Router(path='/my-iew', route_handlers=[MyView])
        ]

        # routers.py
        from esmerald import Router
        from esmerald.conf.urls import include

        router_patterns = [
            Router(path='/', routes=[include('myapp.routes', pattern='myapp_urls')])
        ]


    Args:
        namespace (Any): Example: 'mymodule.urls'
        pattern (Optional[str], optional): The name of the list to be read from the
        module. Defaults to `router_patterns`.
    """

    if not isinstance(arg, str):
        raise ImproperlyConfigured(
            "The value should be a string with the format " "<module>.<file>"
        )

    router_conf_module = import_module(arg)
    pattern = pattern or DEFAULT_PATTERN
    patterns = getattr(router_conf_module, pattern, None)

    if not patterns:
        raise ImproperlyConfigured(
            f"There is no pattern {pattern} found in {arg}. "
            "Are you sure you configured it correctly?"
        )

    if not isinstance(patterns, list):
        raise ImproperlyConfigured(f"{patterns} should be a list and not {type(patterns)}.")

    return patterns



from typing import TYPE_CHECKING, Optional

from esmerald.enums import EncodingType
from esmerald.openapi.schema import create_schema, update_schema_field_info
from openapi_schemas_pydantic.v3_1_0.media_type import MediaType as OpenAPIMediaType
from openapi_schemas_pydantic.v3_1_0.request_body import RequestBody

if TYPE_CHECKING:
    from pydantic.fields import ModelField


def create_request_body(field: "ModelField", create_examples: bool) -> Optional[RequestBody]:
    """
    Gets the request body of the handler.
    """
    media_type = field.field_info.extra.get("media_type", EncodingType.JSON)
    schema = create_schema(field=field, create_examples=create_examples)
    update_schema_field_info(schema=schema, field_info=field.field_info)
    return RequestBody(
        required=True, content={media_type: OpenAPIMediaType(media_type_schema=schema)}
    )



from typing import TYPE_CHECKING, Any, Dict, Optional, Type, cast

from esmerald.parsers import BaseModelExtra, HashableBaseModel
from esmerald.transformers.datastructures import Signature
from esmerald.typing import Void
from esmerald.utils.helpers import is_async_callable

if TYPE_CHECKING:
    from pydantic.typing import AnyCallable, DictAny


class Inject(HashableBaseModel, BaseModelExtra):
    def __init__(self, dependency: "AnyCallable", use_cache: bool = False, **kwargs: "DictAny"):
        super().__init__(**kwargs)
        self.dependency = cast("AnyCallable", dependency)
        self.signature_model: Optional["Type[Signature]"] = None
        self.use_cache = use_cache
        self.value: Any = Void

    async def __call__(self, **kwargs: Dict[str, Any]) -> Any:
        if self.use_cache and self.value is not Void:
            return self.value

        if is_async_callable(self.dependency):
            value = await self.dependency(**kwargs)
        else:
            value = self.dependency(**kwargs)

        if self.use_cache:
            self.value = value

        return value

    def __eq__(self, other: Any) -> bool:
        return other is self or (
            isinstance(other, self.__class__)
            and other.dependency == self.dependency
            and other.use_cache == self.use_cache
            and other.value == self.value
        )



"""Esmerald permission system"""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from esmerald.exceptions import ImproperlyConfigured

if TYPE_CHECKING:
    from esmerald.requests import Request
    from esmerald.types import APIGateHandler

SAFE_METHODS = ("GET", "HEAD", "OPTIONS")


class BaseOperationHolder:
    def __and__(self, other):
        return OperandHolder(AND, self, other)

    def __or__(self, other):
        return OperandHolder(OR, self, other)

    def __rand__(self, other):
        return OperandHolder(AND, other, self)

    def __ror__(self, other):
        return OperandHolder(OR, other, self)

    def __invert__(self):
        return SingleOperand(NOT, self)


class SingleOperand(BaseOperationHolder):
    def __init__(self, operator_class, op1_class):
        self.operator_class = operator_class
        self.op1_class = op1_class

    def __call__(self, *args, **kwargs):
        op1 = self.op1_class(*args, **kwargs)
        return self.operator_class(op1)


class OperandHolder(BaseOperationHolder):
    def __init__(self, operator_class, op1_class, op2_class):
        self.operator_class = operator_class
        self.op1_class = op1_class
        self.op2_class = op2_class

    def __call__(self, *args, **kwargs):
        op1 = self.op1_class(*args, **kwargs)
        op2 = self.op2_class(*args, **kwargs)
        return self.operator_class(op1, op2)


class AND:
    def __init__(self, op1, op2):
        self.op1 = op1
        self.op2 = op2

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return self.op1.has_permission(request, apiview) and self.op2.has_permission(
            request, apiview
        )


class OR:
    def __init__(self, op1, op2):
        self.op1 = op1
        self.op2 = op2

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return self.op1.has_permission(request, apiview) or self.op2.has_permission(
            request, apiview
        )


class NOT:
    def __init__(self, op1):
        self.op1 = op1

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return not self.op1.has_permission(request, apiview)


class BasePermissionMetaclass(BaseOperationHolder, type):
    ...


class BasePermission(metaclass=BasePermissionMetaclass):
    """
    A base class from which all permission classes should inherit.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        return True


class BaseAbstractUserPermission(ABC):
    """
    Abstract Base class for user validation permissions.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        try:
            hasattr(request, "user")
        except ImproperlyConfigured:
            return False

    @abstractmethod
    def is_user_authenticated(self, request: "Request") -> bool:
        """
        This method must be overridden by subclasses.

        Args:
            request: A Starlette 'HTTPConnection' instance.

        Returns:
            bool: True or False
        """
        raise NotImplementedError("is_user_uthenticated() must be implemented.")

    @abstractmethod
    def is_user_staff(self, request: "Request") -> bool:
        """
        This method must be overridden by subclasses.

        Args:
            request: A Starlette 'HTTPConnection' instance.

        Returns:
            bool: True or False
        """
        raise NotImplementedError("is_user_staff() must be implemented.")


class AllowAny(BasePermission):
    """
    Allow any access.
    This isn't strictly required, since you could use an empty
    permission_classes list, but it's useful because it makes the intention
    more explicit.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return True


class DenyAll(BasePermission):
    """
    Deny all access.
    This isn't strictly required, since you could use an empty
    permission_classes list, but it's useful because it makes the intention
    more explicit.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        return False


class IsAuthenticated(BaseAbstractUserPermission):
    """
    Allows access only to authenticated users.
    Raises exception if the AuthenticationMiddleware is not in the `middleware` settings.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        """
        Args:
            request: A Starlette 'HTTPConnection' instance.
            apiview: A Esmerald 'APIController' instance or a `APIGateHandler` instance.

        Returns:
            bool: True or False
        """
        super().has_permission(request, apiview)
        return bool(request.user and self.is_user_authenticated(request))


class IsAdminUser(BaseAbstractUserPermission):
    """
    Allows access only to admin users.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        """
        Args:
            request: A Starlette 'HTTPConnection' instance.
            apiview: A Esmerald 'APIController' instance or a `APIGateHandler` instance.

        Returns:
            bool: True or False
        """
        super().has_permission(request, apiview)
        return bool(request.user and self.is_user_staff(request))


class IsAuthenticatedOrReadOnly(BaseAbstractUserPermission):
    """
    The request is authenticated as a user, or is a read-only request.
    """

    def has_permission(
        self,
        request: "Request",
        apiview: "APIGateHandler",
    ) -> bool:
        """
        Args:
            request: A Starlette 'HTTPConnection' instance.
            apiview: A Esmerald 'APIController' instance or a `APIGateHandler` instance.

        Returns:
            bool: True or False
        """
        super().has_permission(request, apiview)
        return bool(
            request.method in SAFE_METHODS or request.user and self.is_user_authenticated(request)
        )



from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, DirectoryPath, constr, validator
from starlette.staticfiles import StaticFiles

from esmerald.utils.url import clean_path

if TYPE_CHECKING:
    from starlette.types import ASGIApp


class StaticFilesConfig(BaseModel):
    path: constr(min_length=1)  # type: ignore
    directory: Optional[DirectoryPath] = None
    html: bool = False
    packages: Optional[List[Union[str, Tuple[str, str]]]] = None
    check_dir: bool = True

    @validator("path")
    def validate_path(cls, value: str) -> str:  # pylint: disable=no-self-argument
        if "{" in value:
            raise ValueError("path parameters are not supported for static files")
        return clean_path(value)

    def _build_kwargs(
        self,
    ) -> Dict[str, Union[bool, int, DirectoryPath, List[Union[str, Tuple[str, str]]]]]:
        """
        Builds the necessary kwargs to create an StaticFiles object.
        """
        kwargs = {"html": self.html, "check_dir": self.check_dir}
        if self.packages:
            kwargs.update({"packages": self.packages})
        if self.directory:
            kwargs.update({"directory": str(self.directory)})
        return kwargs

    def to_app(self) -> "ASGIApp":
        """
        It can be three scenarios
        """

        return StaticFiles(**self._build_kwargs())



from typing import (
    TYPE_CHECKING,
    Any,
    List,
    NamedTuple,
    Set,
    Tuple,
    Type,
    cast,
)

from esmerald.enums import ParamType, ScopeType
from esmerald.exceptions import ImproperlyConfigured, ValidationErrorException
from esmerald.parsers import BaseModelExtra, HashableBaseModel
from esmerald.requests import Request
from esmerald.utils.constants import REQUIRED
from pydantic.fields import FieldInfo, Undefined
from starlette.datastructures import URL

if TYPE_CHECKING:
    from esmerald.injector import Inject
    from esmerald.transformers.datastructures import EsmeraldSignature, Parameter
    from esmerald.typing import ConnectionType
    from pydantic.typing import DictAny, MappingIntStrAny


class ParamSetting(NamedTuple):
    default_value: Any
    field_alias: str
    field_name: str
    is_required: bool
    is_sequence: bool
    param_type: ParamType
    field_info: FieldInfo


class Dependency(HashableBaseModel, BaseModelExtra):
    def __init__(
        self, key: str, inject: "Inject", dependencies: List["Dependency"], **kwargs: "DictAny"
    ) -> None:
        super().__init__(**kwargs)
        self.key = key
        self.inject = inject
        self.dependencies = dependencies

    class Config(BaseModelExtra.Config):
        arbitrary_types_allowed = True


def merge_sets(first_set: Set[ParamSetting], second_set: Set[ParamSetting]) -> Set[ParamSetting]:
    merged_result = first_set.intersection(second_set)
    difference = first_set.symmetric_difference(second_set)
    for parameter in difference:
        if parameter.is_required or not any(
            param.field_alias == parameter.field_alias and param.is_required
            for param in difference
        ):
            merged_result.add(parameter)
    return merged_result


def create_parameter_setting(
    allow_none: bool,
    field_info: FieldInfo,
    field_name: str,
    path_parameters: Set[str],
    is_sequence: bool,
) -> ParamSetting:
    """
    Creates a setting definition for a parameter.
    """
    extra = field_info.extra
    is_required = extra.get(REQUIRED, True)
    default_value = field_info.default if field_info.default is not Undefined else None

    field_alias = extra.get(ParamType.QUERY) or field_name
    param_type = getattr(field_info, "in_", ParamType.QUERY)

    if field_name in path_parameters:
        field_alias = field_name
        param_type = param_type.PATH
    elif extra.get(ParamType.HEADER):
        field_alias = extra[ParamType.HEADER]
        param_type = ParamType.HEADER
    elif extra.get(ParamType.COOKIE):
        field_alias = extra[ParamType.COOKIE]
        param_type = ParamType.COOKIE

    param_settings = ParamSetting(
        param_type=param_type,
        field_alias=field_alias,
        default_value=default_value,
        field_name=field_name,
        field_info=field_info,
        is_sequence=is_sequence,
        is_required=is_required and (default_value is None and not allow_none),
    )
    return param_settings


def get_request_params(
    params: "MappingIntStrAny", expected: Set[ParamSetting], url: URL
) -> "DictAny":
    """
    Gather the parameters from the request.
    """
    _params = []
    for param in expected:
        if param.is_required and param.field_alias not in params:
            _params.append(param.field_alias)
    if _params:
        raise ValidationErrorException(
            f"Missing required parameter(s) {', '.join(_params)} for url {url}."
        )
    values = {
        param.field_name: params.get(param.field_alias, param.default_value) for param in expected
    }
    return values


def get_connection_info(connection: "ConnectionType") -> Tuple[str, "URL"]:
    """
    Extacts the information from the ConnectionType.
    """
    method = connection.method if isinstance(connection, Request) else ScopeType.WEBSOCKET
    return method, connection.url


def get_signature(value: Any) -> Type["EsmeraldSignature"]:
    try:
        return cast("Type[EsmeraldSignature]", getattr(value, "signature_model"))
    except AttributeError as exc:
        raise ImproperlyConfigured(f"The 'signature' attribute for {value} is not set.") from exc


def get_field_definition_from_param(param: "Parameter") -> Tuple[Any, Any]:
    if param.default_defined:
        definition = (param.annotation, param.default)
    elif not param.optional:
        definition = (param.annotation, ...)
    else:
        definition = (param.annotation, None)
    return definition



from typing import Optional, Set

from pydantic import BaseModel
from typing_extensions import Literal

from esmerald.types import HTTPMethod


class CSRFConfig(BaseModel):
    secret: str
    cookie_name: str = "csrftoken"
    cookie_path: str = "/"
    header_name: str = "X-CSRFToken"
    cookie_secure: bool = False
    cookie_httponly: bool = False
    cookie_samesite: Literal["lax", "strict", "none"] = "lax"
    cookie_domain: Optional[str] = None
    safe_methods: Set[HTTPMethod] = {"GET", "HEAD"}



from starlette.middleware.base import BaseHTTPMiddleware  # noqa
from starlette.middleware.base import (
    RequestResponseEndpoint as RequestResponseEndpoint,
)  # noqa

from esmerald.requests import Request
from esmerald.responses import Response


class BasicHTTPMiddleware(BaseHTTPMiddleware):
    """
    BaseHTTPMiddleware of all Esmerald applications.
    """

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        response = await call_next(request)
        return response



from abc import ABC
from typing import TYPE_CHECKING

from esmerald.protocols.interceptor import InterceptorProtocol

if TYPE_CHECKING:
    from starlette.types import Receive, Scope, Send


class EsmeraldInterceptor(ABC, InterceptorProtocol):
    """Base class for any Esmerald interceptor in the system."""

    async def intercept(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        The abstract method that needs to be implemented for any interceptor.
        """
        raise NotImplementedError("intercept must be implemented")



import hashlib
import random
import time

from esmerald.conf import settings

try:
    random = random.SystemRandom()
    using_sysrandom = True
except NotImplementedError:
    import warnings

    warnings.warn(
        "A secure pseudo-random number generator is not available "
        "on your system. Falling back to Mersenne Twister."
    )
    using_sysrandom = False


def get_random_string(
    length: int = 12,
    allowed_chars: str = "abcdefghijklmnopqrstuvwxyz" "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789",
) -> str:
    """
    Returns a securely generated random string.
    The default length of 12 with the a-z, A-Z, 0-9 character set returns
    a 71-bit value. log_2((26+26+10)^12) =~ 71 bits
    """
    if not using_sysrandom:
        # This is ugly, and a hack, but it makes things better than
        # the alternative of predictability. This re-seeds the PRNG
        # using a value that is hard for an attacker to predict, every
        # time a random string is required. This may change the
        # properties of the chosen random sequence slightly, but this
        # is better than absolute predictability.
        random.seed(
            hashlib.sha256(
                ("%s%s%s" % (random.getstate(), time.time(), settings.secret_key)).encode("utf-8")
            ).digest()
        )
    return "".join(random.choice(allowed_chars) for i in range(length))


def get_random_secret_key():
    """
    Return a 50 character random string usable as a SECRET_KEY setting value.
    """
    chars = "abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)"
    return get_random_string(32, chars)



"""
The MIT License (MIT)

Copyright (c) 2021, 2022 Starlite-API

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import hashlib
import hmac
import secrets
from typing import TYPE_CHECKING, Optional

from esmerald.datastructures import Cookie
from esmerald.enums import ScopeType
from esmerald.exceptions import PermissionDenied
from esmerald.protocols.middleware import MiddlewareProtocol
from esmerald.requests import Request
from starlette.datastructures import MutableHeaders

if TYPE_CHECKING:
    from esmerald.config import CSRFConfig
    from starlette.types import ASGIApp, Message, Receive, Scope, Send

CSRF_SECRET_BYTES = 32
CSRF_SECRET_LENGTH = CSRF_SECRET_BYTES * 2


class CSRFMiddleware(MiddlewareProtocol):
    def __init__(
        self,
        app: "ASGIApp",
        config: "CSRFConfig",
    ):
        """CSRF Middleware class.

        This Middleware protects against attacks by setting a CSRF cookie with a token and verifying it in request headers.

        Args:
            app: The 'next' ASGI app to call.
            config: The CSRFConfig instance.
        """
        super().__init__(app)
        self.app = app
        self.config = config

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        if scope["type"] != ScopeType.HTTP:
            await self.app(scope, receive, send)
            return

        request = Request(scope=scope)
        csrf_cookie = request.cookies.get(self.config.cookie_name)
        current_token = request.headers.get(self.config.header_name)

        if request.method in self.config.safe_methods:
            await self.app(scope, receive, self.get_send_wrapper(send, csrf_cookie))
        elif self._csrf_tokens_match(current_token, csrf_cookie):
            await self.app(scope, receive, send)
        else:
            raise PermissionDenied(detail="CSRF token verification failed.")

    def get_send_wrapper(self, send: "Send", csrf_cookie: Optional[str]) -> "Send":
        async def send_wrapper(message: "Message") -> None:
            """Send function that wraps the original send to inject a
            cookie.

            Args:
                message: An ASGI 'Message'

            Returns:
                None
            """
            if csrf_cookie is None and message["type"] == "http.response.start":
                message.setdefault("headers", [])
                self._set_cookie_if_needed(message)
            await send(message)

        return send_wrapper

    def _set_cookie_if_needed(self, message: "Message") -> None:
        headers = MutableHeaders(scope=message)
        if "set-cookie" not in headers:
            cookie = Cookie(
                key=self.config.cookie_name,
                value=self._generate_csrf_token(),
                path=self.config.cookie_path,
                secure=self.config.cookie_secure,
                httponly=self.config.cookie_httponly,
                samesite=self.config.cookie_samesite,
                domain=self.config.cookie_domain,
            )
            headers.append("set-cookie", cookie.to_header(header=""))

    def _generate_csrf_hash(self, token: str) -> str:
        """Generate an HMAC that signs the CSRF token."""
        return hmac.new(self.config.secret.encode(), token.encode(), hashlib.sha256).hexdigest()

    def _generate_csrf_token(self) -> str:
        """Generate a CSRF token that includes a randomly generated string
        signed by an HMAC."""
        token = secrets.token_hex(CSRF_SECRET_BYTES)
        token_hash = self._generate_csrf_hash(token)
        return token + token_hash

    def _decode_csrf_token(self, token: str) -> Optional[str]:
        """Decode a CSRF token and validate its HMAC."""
        if len(token) < CSRF_SECRET_LENGTH + 1:
            return None

        token_secret = token[:CSRF_SECRET_LENGTH]
        existing_hash = token[CSRF_SECRET_LENGTH:]
        expected_hash = self._generate_csrf_hash(token_secret)
        if not secrets.compare_digest(existing_hash, expected_hash):
            return None

        return token_secret

    def _csrf_tokens_match(
        self, request_csrf_token: Optional[str], cookie_csrf_token: Optional[str]
    ) -> bool:
        """Takes the CSRF tokens from the request and the cookie and verifies
        both are valid and identical."""
        if not (request_csrf_token and cookie_csrf_token):
            return False

        decoded_request_token = self._decode_csrf_token(request_csrf_token)
        decoded_cookie_token = self._decode_csrf_token(cookie_csrf_token)
        if decoded_request_token is None or decoded_cookie_token is None:
            return False

        return secrets.compare_digest(decoded_request_token, decoded_cookie_token)



from tortoise import fields
from tortoise.models import Model

from esmerald.contrib.auth.hashers import (
    check_password,
    is_password_usable,
    make_password,
)


class AutoIncrementIntMixin(Model):
    """
    Auto increment for integers.
    """

    id = fields.IntField(pk=True)

    class Meta:
        ordering = ["-id"]
        abstract = True


class AutoIncrementBigIntMixin(Model):
    """
    Auto increment for big integers.
    """

    id = fields.BigIntField(pk=True)

    class Meta:
        ordering = ["-id"]
        abstract = True


class AbstractUser(Model):
    """
    Base model used for a custom model of an application contains.
    """

    first_name = fields.CharField(description="First name", max_length=150)
    last_name = fields.CharField(description="Last name", max_length=150)
    username = fields.CharField(description="Username", max_length=150, unique=True)
    email = fields.CharField(description="Email address", max_length=120, unique=True)
    password = fields.CharField(description="Password", max_length=128)
    last_login = fields.DatetimeField(description="Last login", null=True)
    is_active = fields.BooleanField(default=True)
    is_staff = fields.BooleanField(default=False)
    is_superuser = fields.BooleanField(default=False)

    # Stores the raw password if set_password() is called so that it can
    # be passed to password_changed() after the model is saved.
    _password = None

    class Meta:
        abstract = True

    @property
    async def is_authenticated(self):
        """
        Always return True.
        """
        return True

    async def set_password(self, raw_password):
        self.password = make_password(raw_password)
        self._password = raw_password
        await self.save()

    async def check_password(self, raw_password):
        """
        Return a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """

        async def setter(raw_password):
            self.set_password(raw_password)
            # Password hash upgrades shouldn't be considered password changes.
            self._password = None
            await self.save(update_fields=["password"])

        return check_password(raw_password, self.password, setter)

    async def set_unusable_password(self):
        # Set a value that will never be a valid hash
        self.password = make_password(None)

    async def has_usable_password(self):
        """
        Return False if set_unusable_password() has been called for this user.
        """
        return is_password_usable(self.password)

    @classmethod
    async def _create_user(cls, username, email, password, **extra_fields):
        """
        Create and save a user with the given username, email, and password.
        """
        if not username:
            raise ValueError("The given username must be set")
        user = cls(username=username, email=email, **extra_fields)
        user.password = make_password(password)
        await user.save()
        return user

    @classmethod
    async def create_user(cls, username, email=None, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        return await cls._create_user(username, email, password, **extra_fields)

    @classmethod
    async def create_superuser(cls, username, email=None, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)

        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")

        return await cls._create_user(username, email, password, **extra_fields)


class User(AbstractUser):
    """
    Implementation of the Abstract user and can be used directly.
    """

    ...



from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from functools import partial
from inspect import Signature, isawaitable
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

from starlette.convertors import CONVERTOR_TYPES
from starlette.requests import HTTPConnection
from starlette.responses import JSONResponse
from starlette.responses import Response as StarletteResponse
from starlette.routing import Mount as Mount  # noqa
from starlette.routing import compile_path
from starlette.types import Receive, Scope, Send
from typing_extensions import TypedDict

from esmerald.backgound import BackgroundTask, BackgroundTasks
from esmerald.datastructures import ResponseContainer
from esmerald.enums import MediaType
from esmerald.exceptions import ImproperlyConfigured
from esmerald.injector import Inject
from esmerald.permissions.utils import continue_or_raise_permission_exception
from esmerald.requests import Request
from esmerald.responses import ORJSONResponse, Response, UJSONResponse
from esmerald.routing.views import APIView
from esmerald.transformers.model import TransformerModel
from esmerald.transformers.signature import SignatureFactory
from esmerald.transformers.utils import get_signature
from esmerald.typing import Void
from esmerald.utils.helpers import is_async_callable, is_class_and_subclass
from esmerald.utils.sync import AsyncCallable

if TYPE_CHECKING:
    from pydantic.typing import AnyCallable

    from esmerald.applications import Esmerald
    from esmerald.interceptors.types import Interceptor
    from esmerald.permissions.types import Permission
    from esmerald.routing.router import HTTPHandler, WebSocketHandler
    from esmerald.types import (
        AsyncAnyCallable,
        ExceptionHandlers,
        ParentType,
        ResponseCookies,
        ResponseHeaders,
        ResponseType,
    )

param_type_map = {
    "str": str,
    "int": int,
    "float": float,
    "uuid": UUID,
    "decimal": Decimal,
    "date": date,
    "datetime": datetime,
    "time": time,
    "timedelta": timedelta,
    "path": Path,
}

CONV2TYPE = {conv: typ for typ, conv in CONVERTOR_TYPES.items()}


T = TypeVar("T", bound="BaseHandlerMixin")


class PathParameterSchema(TypedDict):
    name: str
    full: str
    type: Type


class OpenAPIDefinitionMixin:
    def parse_path(self, path: str) -> Tuple[str, str, List[Union[str, PathParameterSchema]]]:
        """
        Using the Starlette CONVERTORS and the application registered convertors,
        transforms the path into a PathParameterSchema used for the OpenAPI definition.
        """
        _, path, variables = compile_path(path)

        parsed_components: List[Union[str, PathParameterSchema]] = []

        for name, convertor in variables.items():
            _type = CONV2TYPE[convertor]
            parsed_components.append(
                PathParameterSchema(name=name, type=param_type_map[_type], full=name)
            )
        return parsed_components


class BaseSignature:
    """
    In charge of handling the signartures of the handlers.
    """

    def create_signature_model(self, is_websocket: bool = False) -> None:
        """
        Creates a signature model for the given route.

        Websockets do not support methods.
        """
        if not self.signature_model:
            self.signature_model = SignatureFactory(
                fn=cast("AnyCallable", self.fn),
                dependency_names=self.dependency_names,
            ).create_signature()

        for dependency in list(self.get_dependencies().values()):
            if not dependency.signature_model:
                dependency.signature_model = SignatureFactory(
                    fn=dependency.dependency, dependency_names=self.dependency_names
                ).create_signature()

        transformer_model = self.create_handler_transformer_model()
        if not is_websocket:
            self.transformer = transformer_model
            for method in self.methods:
                self.route_map[method] = (self, transformer_model)
        else:
            self.websocket_parameter_model = transformer_model

    def create_handler_transformer_model(self) -> "TransformerModel":
        """Method to create a TransformerModel for a given handler."""
        dependencies = self.get_dependencies()
        signature_model = get_signature(self)

        return TransformerModel.create_signature(
            signature_model=signature_model,
            dependencies=dependencies,
            path_parameters=self.path_parameters,
        )


class BaseResponseHandler:
    """
    In charge of handling the responses of the handlers.
    """

    def response_container_handler(
        self,
        cookies: "ResponseCookies",
        headers: Dict[str, Any],
        media_type: str,
        status_code: int,
    ) -> "AsyncAnyCallable":
        """Creates a handler for ResponseContainer Types"""

        async def response_content(
            data: ResponseContainer, app: "Esmerald", **kwargs: Dict[str, Any]
        ) -> StarletteResponse:
            _headers = {**self.get_headers(headers), **data.headers}
            _cookies = self.get_cookies(data.cookies, cookies)
            response = data.to_response(
                app=app,
                headers=_headers,
                status_code=status_code,
                media_type=media_type,
            )
            for cookie in _cookies:
                response.set_cookie(**cookie)
            return response

        return response_content

    def response_handler(
        self,
        cookies: "ResponseCookies",
        headers: Optional["ResponseHeaders"] = None,
        status_code: Optional[int] = None,
        media_type: Optional[str] = MediaType.TEXT,
    ) -> "AsyncAnyCallable":
        async def response_content(data: Response, **kwargs: Dict[str, Any]) -> StarletteResponse:
            _cookies = self.get_cookies(data.cookies, cookies)
            _headers = {
                **self.get_headers(headers),
                **data.headers,
                **self.allow_header,
            }

            for cookie in _cookies:
                data.set_cookie(**cookie)

            if status_code:
                data.status_code = status_code

            if media_type:
                data.media_type = media_type

            for header, value in _headers.items():
                data.headers[header] = value
            return data

        return response_content

    def json_response_handler(
        self,
        status_code: Optional[int] = None,
        cookies: Optional["ResponseCookies"] = None,
        headers: Optional["ResponseHeaders"] = None,
    ) -> "AsyncAnyCallable":
        """Creates a handler function for Esmerald JSON responses"""

        async def response_content(data: Response, **kwargs: Dict[str, Any]) -> StarletteResponse:
            _cookies = self.get_cookies(cookies, [])
            _headers = {
                **self.get_headers(headers),
                **data.headers,
                **self.allow_header,
            }
            for cookie in _cookies:
                data.set_cookie(**cookie)

            for header, value in _headers.items():
                data.headers[header] = value

            if status_code:
                data.status_code = status_code
            return data

        return response_content

    def starlette_response_handler(
        self,
        cookies: "ResponseCookies",
        headers: Optional["ResponseHeaders"] = None,
        media_type: Optional["MediaType"] = MediaType.TEXT,
    ) -> "AsyncAnyCallable":
        """Creates an handler for Starlette Responses."""

        async def response_content(
            data: StarletteResponse, **kwargs: Dict[str, Any]
        ) -> StarletteResponse:

            _cookies = self.get_cookies(cookies, [])
            _headers = {
                **self.get_headers(headers),
                **data.headers,
                **self.allow_header,
            }
            for cookie in _cookies:
                data.set_cookie(**cookie)

            data.media_type = media_type

            for header, value in _headers.items():
                data.headers[header] = value
            return data  # type: ignore

        return response_content

    def handler(
        self,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]],
        cookies: "ResponseCookies",
        headers: Dict[str, Any],
        media_type: str,
        response_class: "ResponseType",
        status_code: int,
    ) -> "AsyncAnyCallable":
        async def response_content(data: Any, **kwargs: Dict[str, Any]) -> StarletteResponse:

            data = await self.get_response_data(data=data)
            _cookies = self.get_cookies(cookies, [])
            # Making sure ORJSONResponse and JSONResponse are properly handled
            if isinstance(data, (JSONResponse, ORJSONResponse, UJSONResponse)):
                response = data
                response.status_code = status_code
                response.background = background
            else:
                response = response_class(
                    background=background,
                    content=data,
                    headers=headers,
                    media_type=media_type,
                    status_code=status_code,
                )

            for cookie in _cookies:
                response.set_cookie(**cookie)
            return response  # type: ignore

        return response_content

    async def get_response_for_request(
        self,
        scope: "Scope",
        request: Request,
        route: Union["HTTPHandler", "WebSocketHandler"],
        parameter_model: "TransformerModel",
    ) -> "StarletteResponse":
        response: Optional["StarletteResponse"] = None
        if not response:
            response = await self.call_handler_function(
                scope=scope,
                request=request,
                route=route,
                parameter_model=parameter_model,
            )

        return response

    async def call_handler_function(
        self,
        scope: "Scope",
        request: Request,
        route: Union["HTTPHandler", "WebSocketHandler"],
        parameter_model: "TransformerModel",
    ) -> "StarletteResponse":
        response_data = None

        if not response_data:
            response_data = await self._get_response_data(
                route=route,
                parameter_model=parameter_model,
                request=request,
            )

        return await self.to_response(
            app=scope["app"],
            data=response_data,
        )

    @staticmethod
    async def _get_response_data(
        route: "HTTPHandler", parameter_model: "TransformerModel", request: Request
    ) -> Any:
        """
        Determines what kwargs are required for the given handler and assignes to the specific
        object dictionary.

        It supports more one object payload to be sent.
        """
        signature_model = get_signature(route)

        if parameter_model.has_kwargs:
            kwargs = parameter_model.to_kwargs(connection=request)
            request_data = kwargs.get("data")
            if request_data:
                kwargs["data"] = await request_data
            for dependency in parameter_model.dependencies:
                kwargs[dependency.key] = await parameter_model.get_dependencies(
                    dependency=dependency, connection=request, **kwargs
                )
            parsed_kwargs = signature_model.parse_values_for_connection(
                connection=request, **kwargs
            )
        else:
            parsed_kwargs = {}
        if isinstance(route.parent, APIView):
            fn = partial(
                cast("AnyCallable", route.fn),
                route.parent,
                **parsed_kwargs,
            )
        else:
            fn = partial(cast("AnyCallable", route.fn), **parsed_kwargs)

        if is_async_callable(fn):
            return await fn()
        return fn()

    def get_response_handler(self) -> Callable[[Any], Awaitable[StarletteResponse]]:
        """
        Checks and validates the type of return response and maps to the corresponding
        handler with the given parameters.
        """
        if self._response_handler is Void:
            media_type = (
                self.media_type.value if isinstance(self.media_type, Enum) else self.media_type
            )

            response_class = self.get_response_class()
            headers = self.get_response_headers()
            cookies = self.get_response_cookies()

            if is_class_and_subclass(self.signature.return_annotation, ResponseContainer):
                handler = self.response_container_handler(
                    cookies=cookies,
                    media_type=self.media_type,
                    status_code=self.status_code,
                    headers=headers,
                )
            elif is_class_and_subclass(
                self.signature.return_annotation,
                (JSONResponse, ORJSONResponse, UJSONResponse),
            ):
                handler = self.json_response_handler(
                    status_code=self.status_code, cookies=cookies, headers=headers
                )
            elif is_class_and_subclass(self.signature.return_annotation, Response):
                handler = self.response_handler(
                    cookies=cookies,
                    status_code=self.status_code,
                    media_type=self.media_type,
                    headers=headers,
                )
            elif is_class_and_subclass(self.signature.return_annotation, StarletteResponse):
                handler = self.starlette_response_handler(
                    cookies=cookies,
                    media_type=self.media_type,
                    headers=headers,
                )
            else:
                handler = self.handler(
                    background=self.background,
                    cookies=cookies,
                    headers=headers,
                    media_type=media_type,
                    response_class=response_class,
                    status_code=self.status_code,
                )
            self._response_handler = handler
        return cast(
            "Callable[[Any], Awaitable[StarletteResponse]]",
            self._response_handler,
        )


class BaseHandlerMixin(BaseSignature, BaseResponseHandler, OpenAPIDefinitionMixin):
    """
    Base of HTTPHandler and WebSocketHandler.
    """

    @property
    def signature(self) -> Signature:
        """The Signature of 'self.fn'."""
        return Signature.from_callable(cast("AnyCallable", self.fn))

    @property
    def path_parameters(self) -> Set[str]:
        """
        Gets the path parameters in a set format.

        Example: {'name', 'id'}
        """
        parameters = set()
        for param_name, _ in self.param_convertors.items():
            parameters.add(param_name)
        return parameters

    @property
    def normalised_path_params(self) -> List[Dict[str, str]]:
        """
        Gets the path parameters in a PathParameterSchema format and it is
        only used for OpenAPI documentation purposes only.
        """
        path_components = self.parse_path(self.path)
        parameters = [component for component in path_components if isinstance(component, dict)]
        return parameters

    @property
    def parent_levels(self) -> List[Union[T, "ParentType"]]:
        """
        Returns the handler from the app down to the route handler.

        Who is the parent of a given layer/level.

        Example:

            app = Esmerald(routes=[
                Include(path='/api/v1', routes=[
                    Gateway(path='/home', handler=home)
                ])
            ])

            1. Parent of Gateway is the Include.
            2. Parent of the Include is the Esmerald router.
            3. Parent of the Esmerald router is the Esmerald app itself.
        """
        levels = []
        current: Any = self
        while current:
            levels.append(current)
            current = current.parent
        return list(reversed(levels))

    @property
    def dependency_names(self) -> Set[str]:
        """A unique set of all dependency names provided in the handlers parent
        levels."""
        level_dependencies = (level.dependencies or {} for level in self.parent_levels)
        return {name for level in level_dependencies for name in level.keys()}

    def get_permissions(self) -> List["Permission"]:
        """
        Returns all the permissions in the handler scope from the ownsership layers.
        """
        if self._permissions is Void:
            self._permissions = []
            for layer in self.parent_levels:
                self._permissions.extend(layer.permissions or [])
            self._permissions = cast(
                "List[Permission]",
                [AsyncCallable(permissions) for permissions in self._permissions],
            )
        return cast("List[Permission]", self._permissions)

    def get_dependencies(self) -> Dict[str, Inject]:
        """
        Returns all dependencies of the handler function's starting from the parent levels.
        """
        if not self.signature_model:
            raise RuntimeError(
                "get_dependencies cannot be called before a signature model has been generated"
            )
        if self._dependencies is Void:
            self._dependencies = {}
            for level in self.parent_levels:
                for key, value in (level.dependencies or {}).items():
                    self.is_unique_dependency(
                        dependencies=self._dependencies,
                        key=key,
                        injector=value,
                    )
                    self._dependencies[key] = value
        return cast("Dict[str, Inject]", self._dependencies)

    @staticmethod
    def is_unique_dependency(dependencies: Dict[str, Inject], key: str, injector: Inject) -> None:
        """
        Validates that a given inject has not been already defined under a
        different key in any of the levels.
        """
        for dependency_key, value in dependencies.items():
            if injector == value:
                raise ImproperlyConfigured(
                    f"Injector for key {key} is already defined under the different key {dependency_key}. "
                    f"If you wish to override a inject, it must have the same key."
                )

    def get_exception_handlers(self) -> "ExceptionHandlers":
        """
        Resolves the exception_handlers by starting from the route handler
        and moving up.
        """
        resolved_exception_handlers = {}
        for level in self.parent_levels:
            resolved_exception_handlers.update(level.exception_handlers or {})
        return resolved_exception_handlers

    def get_cookies(
        self, local_cookies: "ResponseCookies", other_cookies: "ResponseCookies"
    ) -> List[Dict[str, Any]]:
        """
        Returns a unique list of cookies.
        """
        filtered_cookies = [*local_cookies]
        for cookie in other_cookies:
            if not any(cookie.key == c.key for c in filtered_cookies):
                filtered_cookies.append(cookie)
        normalized_cookies: List[Dict[str, Any]] = []
        for cookie in filtered_cookies:
            normalized_cookies.append(cookie.dict(exclude_none=True, exclude={"description"}))
        return normalized_cookies

    def get_headers(self, headers: "ResponseHeaders") -> Dict[str, Any]:
        """
        Returns a dict of response headers.
        """
        return {k: v.value for k, v in headers.items()}

    async def get_response_data(self, data: Any) -> Any:
        """
        Retrives the response data for sync and async.
        """
        if isawaitable(data):
            data = await data
        return data

    async def allow_connection(self, connection: "HTTPConnection") -> None:
        """
        Validates the connection.

        Handles with the permissions for each view (get, put, post, delete, patch, route...) after the request.

        Raises a PermissionDenied exception if not allowed..
        """
        for permission in self.get_permissions():
            permission = await permission()
            permission.has_permission(connection, self)
            continue_or_raise_permission_exception(connection, self, permission)


class BaseInterceptorMixin(BaseHandlerMixin):
    def get_interceptors(self) -> List["Interceptor"]:
        """
        Returns all the interceptors in the handler scope from the ownsership layers.
        """
        if self._interceptors is Void:
            self._interceptors = []
            for layer in self.parent_levels:
                self._interceptors.extend(layer.interceptors or [])
            self._interceptors = cast(
                "List[Interceptor]",
                [AsyncCallable(interceptors) for interceptors in self._interceptors],
            )
        return cast("List[Interceptor]", self._interceptors)

    async def intercept(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Checks for every interceptor on each level and runs them all before reaching any
        of the handlers.
        """
        for interceptor in self.get_interceptors():
            interceptor = await interceptor()
            await interceptor.intercept(scope, receive, send)



"""
Esmerald readapted the same concept to allow a simple integration for the
`esmerald-admin` allowing the creation of a project and app via command line.
"""
import os
import shutil
import stat
from importlib import import_module
from typing import Any, Dict, Union

import esmerald
from esmerald.core.management.base import BaseDirective, DirectiveError
from jinja2 import Environment, FileSystemLoader
from pydantic import FilePath


class TemplateDirective(BaseDirective):
    """
    Copy either an Esmerald application layout template or an Esmerald project
    layout.
    """

    url_schemes = ["http", "https", "ftp"]

    rewrite_template_suffixes = (
        (".py-tpl", ".py"),
        (".e-tpl", ""),
    )

    def add_arguments(self, parser: Any):
        parser.add_argument("name", help="Name of the application or project.")

    def handle(
        self,
        app_or_project: str,
        name: str,
        **options: Dict[str, Any],
    ):
        self.app_or_project = app_or_project
        self.a_or_an = "an" if app_or_project == "app" else "a"
        self.paths_to_remove = []
        self.verbosity = options["verbosity"]

        self.validate_name(name)

        top_dir = os.path.join(os.getcwd(), name)
        try:
            os.makedirs(top_dir)
        except FileExistsError:
            raise DirectiveError(f"{top_dir} already exists.")
        except OSError as e:
            raise DirectiveError(e)

        base_name = f"{app_or_project}_name"
        base_subdir = f"{app_or_project}_template"

        context = {
            base_name: name,
            "esmerald_version": self.get_version(),
            "project_secret": options.get("secret_key"),
        }

        template_dir = os.path.join(esmerald.__path__[0], "conf", base_subdir)
        prefix_length = len(template_dir) + 1

        for root, dirs, files in os.walk(template_dir):
            path_rest = root[prefix_length:]
            relative_dir = path_rest.replace(base_name, name)
            if relative_dir:
                target_dir = os.path.join(top_dir, relative_dir)
                os.makedirs(target_dir, exist_ok=True)

            for dirname in dirs[:]:
                if dirname.startswith(".") or dirname == "__pycache__":
                    dirs.remove(dirname)

            for filename in files:
                if filename.endswith((".pyo", ".pyc", ".py.class")):
                    continue

                old_path = os.path.join(root, filename)
                new_path = os.path.join(top_dir, relative_dir, filename.replace(base_name, name))
                project_dir = os.path.join(top_dir, relative_dir)
                template_name = filename
                for old_suffix, new_suffix in self.rewrite_template_suffixes:
                    if new_path.endswith(old_suffix):
                        new_path = new_path[: -len(old_suffix)] + new_suffix
                        template_name = template_name[: -len(old_suffix)] + new_suffix
                        break  # Only rewrite once

                if os.path.exists(new_path):
                    raise DirectiveError(
                        "%s already exists. Overlaying %s %s into an existing "
                        "directory won't replace conflicting files."
                        % (
                            new_path,
                            self.a_or_an,
                            app_or_project,
                        )
                    )
                shutil.copyfile(old_path, new_path)
                if self.verbosity >= 2:
                    self.stdout.write("Creating %s" % new_path)
                try:
                    self.manage_template_variables(template_name, new_path, project_dir, context)
                    self.apply_umask(old_path, new_path)
                    self.make_file_writable(new_path)
                except OSError:
                    self.stderr.write(
                        "Notice: Couldn't set permission bits on %s. You're "
                        "probably using an uncommon filesystem setup. No "
                        "problem." % new_path,
                        self.style.NOTICE,
                    )

    def manage_template_variables(
        self,
        template: Union[str, FilePath],
        destination: Union[str, FilePath],
        template_dir: Union[str, FilePath],
        context: Dict[str, Any],
    ):
        """
        Goes through every file generated and replaces the variables with the given
        context variables.
        """
        environment = Environment(loader=FileSystemLoader(template_dir))
        template = environment.get_template(template)
        rendered_template = template.render(context)
        if os.path.isfile(destination):
            os.unlink(destination)
        with open(destination, "w") as f:
            f.write(rendered_template)

    def validate_name(self, name, name_or_dir="name"):
        if name is None:
            raise DirectiveError(
                "you must provide {an} {app} name".format(
                    an=self.a_or_an,
                    app=self.app_or_project,
                )
            )
        # Check it's a valid directory name.
        if not name.isidentifier():
            raise DirectiveError(
                "'{name}' is not a valid {app} {type}. Please make sure the "
                "{type} is a valid identifier.".format(
                    name=name,
                    app=self.app_or_project,
                    type=name_or_dir,
                )
            )
        try:
            import_module(name)
        except ImportError:
            pass
        else:
            raise DirectiveError(
                "'{name}' conflicts with the name of an existing Python "
                "module and cannot be used as {an} {app} {type}. Please try "
                "another {type}.".format(
                    name=name,
                    an=self.a_or_an,
                    app=self.app_or_project,
                    type=name_or_dir,
                )
            )

    def apply_umask(self, old_path: Union[str, FilePath], new_path: Union[str, FilePath]):
        current_umask = os.umask(0)
        os.umask(current_umask)
        current_mode = stat.S_IMODE(os.stat(old_path).st_mode)
        os.chmod(new_path, current_mode & ~current_umask)

    def make_file_writable(self, filename: str):
        """
        Make sure that the file is writeable.
        Useful if our source is read-only.
        """
        if not os.access(filename, os.W_OK):
            st = os.stat(filename)
            new_permissions = stat.S_IMODE(st.st_mode) | stat.S_IWUSR
            os.chmod(filename, new_permissions)
